package com.io.yutian.auchestshop.manager;

import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.io.yutian.auchestshop.AuChestShop;
import com.io.yutian.auchestshop.database.SQLIO;
import com.io.yutian.auchestshop.shop.Shop;
import com.io.yutian.auchestshop.shop.ShopChunk;
import com.io.yutian.auchestshop.shop.ShopLocation;
import com.io.yutian.auchestshop.shop.ShopType;
import com.io.yutian.auchestshop.util.ShopUtil;
import org.apache.commons.lang.StringUtils;
import org.bukkit.Chunk;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.block.Block;
import org.bukkit.entity.Player;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;

public class ShopManager {

    private final Map<String, Map<ShopChunk, Map<ShopLocation, Shop>>> shops = Maps.newConcurrentMap();
    private final Set<Shop> loadedShops = Sets.newConcurrentHashSet();

    public void loadAll(SQLIO sqlio) {
        List<Shop> shopList = sqlio.loadAllShops();
        shopList.forEach(this::addShop);
    }

    public void saveAll(SQLIO sqlio) {
        sqlio.saveAllShops(getAllShops());
    }

    public Shop createShop(Player player, Block block, ShopType shopType) {
        Location location = block.getLocation();
        ShopLocation shopLocation = ShopLocation.fromLocation(location);
        Shop shop = new Shop(player.getUniqueId(), shopLocation, shopType);
        addShop(shop);
        AuChestShop.getSQLIO().saveShop(shop);
        return shop;
    }

    public @NotNull List<Shop> getAllShops() {
        List<Shop> shopsCollected = new ArrayList<>();
        for (Map<ShopChunk, Map<ShopLocation, Shop>> shopMapData : getShops().values()) {
            for (Map<ShopLocation, Shop> shopData : shopMapData.values()) {
                shopsCollected.addAll(shopData.values());
            }
        }
        return shopsCollected;
    }

    public @NotNull List<Shop> getAllShops(@NotNull UUID playerUUID) {
        List<Shop> playerShops = new ArrayList<>();
        for (Shop shop : getAllShops()) {
            UUID shopUuid = shop.getOwner();
            if (playerUUID.equals(shopUuid)) {
                playerShops.add(shop);
            }
        }
        return playerShops;
    }


    public @Nullable Shop getShop(UUID shopId) {
        for (Shop shop : getAllShops()) {
            if (shop.getUUID().equals(shopId)) {
                return shop;
            }
        }
        return null;
    }

    public @Nullable Shop getShop(@NotNull Location location) {
        if (!ShopUtil.isShoppables(location.getBlock().getType())) {
            return null;
        }
        ShopChunk shopChunk = ShopChunk.fromLocation(location);
        final Map<ShopLocation, Shop> inChunk = getShops(shopChunk);
        if (inChunk == null) {
            return null;
        }
        ShopLocation shopLocation = ShopLocation.fromLocation(location);
        return inChunk.get(shopLocation);
    }

    public Map<String, Map<ShopChunk, Map<ShopLocation, Shop>>> getShops() {
        return this.shops;
    }

    public Map<ShopLocation, Shop> getShops(@NotNull Chunk chunk) {
        return getShops(chunk.getWorld().getName(), chunk.getX(), chunk.getZ());
    }

    public Map<ShopLocation, Shop> getShops(@NotNull String world, int chunkX, int chunkZ) {
        final Map<ShopChunk, Map<ShopLocation, Shop>> inWorld = this.getShops(world);
        if (inWorld == null) {
            return null;
        }
        return inWorld.get(new ShopChunk(world, chunkX, chunkZ));
    }

    public Map<ShopLocation, Shop> getShops(@NotNull ShopChunk shopChunk) {
        return getShops(shopChunk.getWorld(), shopChunk.getX(), shopChunk.getZ());
    }

    public Map<ShopChunk, Map<ShopLocation, Shop>> getShops(@NotNull String world) {
        return this.shops.get(world);
    }

    public @NotNull List<Shop> getShopsInWorld(@NotNull World world) {
        final List<Shop> worldShops = new ArrayList<>();
        for (final Shop shop : getAllShops()) {
            ShopLocation location = shop.getLocation();
            if (Objects.equals(location.getWorld(), world)) {
                worldShops.add(shop);
            }
        }
        return worldShops;
    }

    public @NotNull List<Shop> getShopsInWorld(@NotNull String worldName) {
        final List<Shop> worldShops = new ArrayList<>();
        for (final Shop shop : getAllShops()) {
            ShopLocation location = shop.getLocation();
            if (StringUtils.equals(worldName, location.getWorld())) {
                worldShops.add(shop);
            }
        }
        return worldShops;
    }

    public void addShop(@NotNull Shop shop) {
        ShopLocation shopLocation = shop.getLocation();
        String world = shopLocation.getWorld();
        Map<ShopChunk, Map<ShopLocation, Shop>> map = shops.computeIfAbsent(world, k -> new HashMap<>());
        int x = (int) Math.floor((shopLocation.getX()) / 16.0);
        int z = (int) Math.floor((shopLocation.getZ()) / 16.0);
        ShopChunk shopChunk = new ShopChunk(world, x, z);
        Map<ShopLocation, Shop> inChunk = map.computeIfAbsent(shopChunk, k -> new HashMap<>());
        inChunk.put(shopLocation, shop);
        loadedShops.add(shop);
    }

    public void removeShop(@NotNull Shop shop) {
        this.loadedShops.remove(shop);
    }

    public Set<Shop> getLoadedShops() {
        return loadedShops;
    }

}
